/*
 * Copyright (c) 2016-2018 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Authors: Adrian Szyndela <adrian.s@samsung.com>
 *	    Łukasz Stelmach <l.stelmach@samsung.com>
 *	    Rafał Pietruch <r.pietruch@samsung.com>
 */

/**
 * @file crash-stack.c
 * @brief This file contains Main module of call stack unwinding program
 *
 * Crash-stack is a single purpose program. Its duty is to show call stack
 * of a crashed program. Crash-stack must be called with proper arguments:
 * PID of a crashed program.
 */
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <json-c/json.h>
#include <limits.h>
#include <linux/prctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/prctl.h>
#include <sys/stat.h>
#include <sys/syscall.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/wait.h>
#include <unistd.h>
#include <assert.h>

#include "crash-stack.h"
#include "report_info.h"

#include "shared/log.h"
#include "shared/util.h"

#define BUF_SIZE (BUFSIZ)
#define HEXA 16
#define PERM_LEN 4
#define ADDR_LEN 33
#define STR_ANONY "[anony]"
#define STR_ANONY_LEN 8
#define INFO_JSON_VERSION 1

#define STRING_FORMAT_SPECIFIER_WITH_MACRO(macro) "%"#macro"s"
#define STR_FS(macro) STRING_FORMAT_SPECIFIER_WITH_MACRO(macro)

// this macro returns the format specifier to capture a string that can
// contain spaces
#define STRINGWW_FORMAT_SPECIFIER_WITH_MACRO(macro) "%"#macro"[^\t\n]"
#define STRWW_FS(macro) STRINGWW_FORMAT_SPECIFIER_WITH_MACRO(macro)

/**
 * @brief definitions for getopt options: identifiers
 */
enum {
	OPT_PID,
	OPT_TID,
	OPT_SIGNUM,
	OPT_OUTPUTFILE,
	OPT_JSONOUTPUTFILE,
	OPT_PRSTATUS_FD
};

/**
 * @brief definitions for getopt options: full specifications
 */
const struct option opts[] = {
	{ "pid", required_argument, 0, OPT_PID },
	{ "tid", required_argument, 0, OPT_TID },
	{ "sig", required_argument, 0, OPT_SIGNUM },
	{ "output", required_argument, 0, OPT_OUTPUTFILE },
	{ "output-json", required_argument, 0, OPT_JSONOUTPUTFILE },
	{ "prstatus_fd", required_argument, 0, OPT_PRSTATUS_FD },
	{ 0, 0, 0, 0 }
};

/**
 * @brief container for information from /proc/PID/maps
 */
struct addr_node {
	uintptr_t startaddr;
	uintptr_t endaddr;
	char perm[PERM_LEN+1];
	char *fpath;
	struct addr_node *next;
};

/* helper functions for reading /proc/PID/maps */
static struct addr_node *get_addr_list_from_maps(int fd);
static void free_all_nodes(struct addr_node *start);
static char *fgets_fd(char *str, int len, int fd);

/**
 * @brief Gets registers information for live process
 *
 * @param pid pid of the live process
 * @return 0 on success, -1 otherwise
 */
static int __get_registers_fd(pid_t pid, int fd)
{
	struct elf_prstatus *prstatus = NULL; ///< NT_PRSTATUS of the failed process
	size_t size = 0;
	char *registers = _crash_stack_get_memory_for_registers(&size);
	int ret = -1;

	if (NULL == registers) {
		_E("Cannot get memory for registers (not implemented for this architecture");
		return -1;
	}

	prstatus = mmap(NULL, sizeof(struct elf_prstatus),
			PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);

	if (prstatus == MAP_FAILED)
		return -1;

	if (prstatus->pr_pid == pid) {
		memcpy(registers, prstatus->pr_reg, size);
		ret = 0;
	}

	munmap(prstatus, sizeof(struct elf_prstatus));

	return ret;
}

/**
 * @brief Print signal number causing dump
 */
static void __crash_stack_print_signal(json_object *jobj, int signo)
{
	if (SIGHUP > signo || signo > SIGSYS) {
		_E("Invalid signal number: %d", signo);
		return;
	}

	json_object_object_add(jobj, K_SIGNAL, json_object_new_int(signo));
}

/**
 * @brief Prints call stack element to the JSON object.
 *
 * @param proc_info gathered call stack element
 * @param jobj JSON object
 */
static void __print_proc_info(ProcInfo *proc_info, json_object *jobj)
{
	char value_str[16+3];
	if (proc_info->name) {
		json_object_object_add(jobj, K_CALL_NAME, json_object_new_string(proc_info->name));
		if (proc_info->offset >= 0) {
			if (sprintf(value_str, "0x%x", proc_info->offset) != -1)
				json_object_object_add(jobj, K_CALL_OFFSET, json_object_new_string(value_str));
			else
				_E("Cannot print offset: %m");
		}
	}

	int res;
	if (sizeof(proc_info->addr) > 4)
		res = sprintf(value_str, "0x%016llx", (long long)proc_info->addr);
	else
		res = sprintf(value_str, "0x%08x", (int32_t)proc_info->addr);

	if (res != -1)
		json_object_object_add(jobj, K_CALL_ADDR, json_object_new_string(value_str));
	else
		_E("Cannot print address: %m");

	json_object_object_add(jobj, K_CALL_M_NAME, json_object_new_string(proc_info->module_name ?: ""));

	if (sprintf(value_str, "0x%x", proc_info->module_offset) != -1)
		json_object_object_add(jobj, K_CALL_M_OFFSET, json_object_new_string(value_str));
	else
		_E("Cannot print module offset: %m");
}

/**
 * @brief Prints call stack to the JSON object.
 *
 * @param callstack gathered call stack database
 * @param jobj JSON object
 * @param pid PID of the live process
 */
static void __print_callstack(Callstack *callstack, json_object *jobj, pid_t pid)
{
	json_object *j_callstack = json_object_new_array();

	for (size_t it = 0; it != callstack->elems; ++it) {
		json_object *j_stack_entry = json_object_new_object();
		__print_proc_info(&callstack->proc[it], j_stack_entry);
		json_object_array_add(j_callstack, j_stack_entry);
	}
	json_object_object_add(jobj, K_CALLSTACK, j_callstack);
}

void callstack_constructor(Callstack *callstack)
{
	size_t it;
	callstack->elems = 0;
	for (it = 0; it < (int)sizeof(callstack->proc)/sizeof(callstack->proc[0]); ++it) {
		callstack->proc[it].offset = -1;
		callstack->proc[it].name = 0;
		callstack->proc[it].module_name = 0;
	}
}

void callstack_destructor(Callstack *callstack)
{
	size_t it;
	for (it = 0; it < callstack->elems; ++it) {
		free(callstack->proc[it].name);
		free(callstack->proc[it].module_name);
	}
}

/**
 * @brief Print full path of executable file
 */
static void __crash_stack_print_exe(json_object *jobj, pid_t pid)
{
	char path[PATH_MAX];
	bool has_path = get_exe_path(pid, path, sizeof(path));

	json_object_object_add(jobj, K_EXE_FILE_PATH, json_object_new_string(has_path ? path : "<unknown>"));
}

/**
 * @brief Print thread information
 *
 * @param jobj JSON object
 * @param pid PID of the inspected process
 * @param tid TID of the inspected thread
 */
static void __crash_stack_print_threads(json_object *jobj, pid_t pid, pid_t tid)
{
	int threadnum = 1;
	DIR *dir;
	struct dirent *dentry = NULL;
	char task_path[PATH_MAX];
	struct stat sb;

	snprintf(task_path, PATH_MAX, "/proc/%d/task", pid);
	if (stat(task_path, &sb) == -1)
		return;

	threadnum = sb.st_nlink - 2;
	json_object *j_threads = json_object_new_object();
	json_object_object_add(j_threads, K_TH_PID, json_object_new_int(pid));
	json_object_object_add(j_threads, K_TH_TID, json_object_new_int(tid));
	json_object *j_tasks = json_object_new_array();

	if (threadnum > 1) {
		/* print thread */
		dir = opendir(task_path);


		if (!dir) {
			_E("opendir(%s): %m", task_path);
		} else {
			while ((dentry = readdir(dir))) {
				if (strcmp(dentry->d_name, ".") == 0 ||
				    strcmp(dentry->d_name, "..") == 0)
					continue;
				json_object_array_add(j_tasks,
				                      json_object_new_int(strtol(dentry->d_name, NULL, 10)));
			}
			closedir(dir);
		}

	} else {
		json_object_array_add(j_tasks, json_object_new_int(tid));
	}
	json_object_object_add(j_threads, K_TH_TASKS, j_tasks);
	json_object_object_add(jobj, K_THREADS, j_threads);
}

/**
 * @brief Print information about mapped memory regions
 *
 * @param jobj JSON object
 * @param pid PID of the inspected process
 */
static void __crash_stack_print_maps(json_object *jobj, pid_t pid)
{
	char file_path[PATH_MAX];
	struct addr_node *head = NULL;
	struct addr_node *t_node;
	int fd;

	snprintf(file_path, PATH_MAX, "/proc/%d/maps", pid);

	if ((fd = open(file_path, O_RDONLY)) < 0) {
		_E("open(%s): %m", file_path);
	} else {
		/* parsing the maps to get code segment address*/
		head = get_addr_list_from_maps(fd);
		close(fd);
	}
	if (head == NULL)
		return;

	t_node = head;
	json_object *j_maps = json_object_new_array();

	while (t_node) {
		if (!strncmp(STR_ANONY, t_node->fpath, STR_ANONY_LEN)) {
			t_node = t_node->next;
		} else {
			char start_str[16+3], end_str[16+3];
			sprintf(start_str, "0x%0*lx", __WORDSIZE/4, (unsigned long)t_node->startaddr);
			sprintf(end_str, "0x%0*lx", __WORDSIZE/4, (unsigned long)t_node->endaddr);

			json_object *j_rec = json_object_new_object();
			json_object_object_add(j_rec, K_MAPS_START, json_object_new_string(start_str));
			json_object_object_add(j_rec, K_MAPS_END, json_object_new_string(end_str));
			json_object_object_add(j_rec, K_MAPS_PERM, json_object_new_string(t_node->perm));
			json_object_object_add(j_rec, K_MAPS_PATH, json_object_new_string(t_node->fpath));
			json_object_array_add(j_maps, j_rec);
			t_node = t_node->next;
		}
	}

	json_object_object_add(jobj, K_MAPS, j_maps);
	free_all_nodes(head);
}

static struct addr_node *get_addr_list_from_maps(int fd)
{
	int fpath_len, result;
	uintptr_t saddr;
	uintptr_t eaddr;
	char perm[PERM_LEN+1];
	char path[PATH_MAX+1];
	char addr[ADDR_LEN+1];
	char linebuf[BUF_SIZE];
	struct addr_node *head = NULL;
	struct addr_node *tail = NULL;
	struct addr_node *t_node = NULL;

	/* parsing the maps to get executable code address */
	while (fgets_fd(linebuf, BUF_SIZE, fd) != NULL) {
		memset(path, 0, PATH_MAX+1);
		result = sscanf(linebuf, STR_FS(ADDR_LEN)
				 STR_FS(PERM_LEN)
				 "%*s %*s %*s "
				 STRWW_FS(PATH_MAX), addr, perm, path);
		if (result < 0)
			continue;
		perm[PERM_LEN] = 0;

		/* rwxp */
		if ((perm[2] == 'x' && path[0] == '/') ||
		    (perm[1] == 'w' && path[0] != '/') ||
		    is_dotnet_file(path)) {
			char* addr2 = strchr(addr, '-');
			if (addr2 == NULL) {
				_E("Not found '-' in addr: %s", addr);
				continue;
			}
			*(addr2++) = '\0';
			/* add addr node to list */
			saddr = strtoul(addr, NULL, HEXA);
			/* ffff0000-ffff1000 */
			eaddr = strtoul(addr2, NULL, HEXA);
			/* make node and attach to the list */
			t_node = (struct addr_node *)mmap(0, sizeof(struct addr_node),
							  PROT_READ | PROT_WRITE,
							  MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
			if (t_node == NULL) {
				_E("mmap(): %m");
				return NULL;
			}
			memcpy(t_node->perm, perm, PERM_LEN+1);
			t_node->startaddr = saddr;
			t_node->endaddr = eaddr;
			t_node->fpath = NULL;
			fpath_len = strlen(path);
			if (fpath_len > 0) {
				t_node->fpath = (char *)mmap(0, fpath_len + 1,
							     PROT_READ | PROT_WRITE,
							     MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
				memset(t_node->fpath, 0, fpath_len + 1);
				memcpy(t_node->fpath, path, fpath_len);
			} else {
				t_node->fpath = (char *)mmap(0, STR_ANONY_LEN,
							     PROT_READ | PROT_WRITE,
							     MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
				memset(t_node->fpath, 0, STR_ANONY_LEN);
				memcpy(t_node->fpath, STR_ANONY, STR_ANONY_LEN);
			}
			t_node->next = NULL;
			if (head == NULL || tail == NULL) {
				assert(head == NULL);
				assert(tail == NULL);
				head = t_node;
				tail = t_node;
			} else {
				tail->next = t_node;
				tail = t_node;
			}
		}
	}
	return head;
}

static void free_all_nodes(struct addr_node *start)
{
	struct addr_node *t_node, *n_node;
	int fpath_len;

	if (start == NULL)
		return;
	t_node = start;
	n_node = t_node->next;
	while (t_node) {
		if (t_node->fpath != NULL) {
			fpath_len = strlen(t_node->fpath);
			munmap(t_node->fpath, fpath_len + 1);
		}
		munmap(t_node, sizeof(struct addr_node));
		if (n_node == NULL)
			break;
		t_node = n_node;
		n_node = n_node->next;
	}
}

static char *fgets_fd(char *str, int len, int fd)
{
	char ch;
	register char *cs;
	int num = 0;

	cs = str;
	while (--len > 0 && (num = read(fd, &ch, 1) > 0)) {
		if ((*cs++ = ch) == '\n')
			break;
	}
	*cs = '\0';
	return (num == 0 && cs == str) ? NULL : str;
}

/**
 * @brief Print process and system memory information
 *
 * @param json_object JSON object
 * @param pid PID of the inspected process
 */
static void __crash_stack_print_meminfo(json_object *jobj, pid_t pid)
{
	char infoname[BUF_SIZE];
	char memsize[BUF_SIZE];
	char linebuf[BUF_SIZE];
	char file_path[PATH_MAX];
	int fd;
	json_object *j_mem_info = json_object_new_object();

#define ADD_TO_JSON_AS_NUM(jobj, name, value) \
	do { \
		errno = 0; \
		long int val = strtol(value, NULL, 10); \
		if ((errno == ERANGE && (val == LONG_MAX || val == LONG_MIN)) || \
		     (errno != 0 && val == 0)) { \
			_E("Can not convert %s (%s) to a numerical value", name, value); \
			val = -1; \
		} \
		json_object_object_add(jobj, name, json_object_new_int(val)); \
	} while (0)


	struct name_name {
		const char *in_file, *in_json;
	} meminfo_names[] = {{"MemTotal:", K_MI_MEMTOTAL},
	                     {"MemFree:",  K_MI_MEMFREE},
	                     {"Buffers:",  K_MI_BUFFS},
	                     {"Cached:",   K_MI_CACHED}};

	if ((fd = open("/proc/meminfo", O_RDONLY)) < 0) {
		_E("open(/proc/meminfo): %m");
	} else {
		while (fgets_fd(linebuf, BUF_SIZE, fd) != NULL) {
			sscanf(linebuf, "%16s %16s %*s", infoname, memsize);
			for (size_t i = 0; i < ARRAY_SIZE(meminfo_names); i++) {
				if (strcmp(meminfo_names[i].in_file, infoname) == 0) {
					ADD_TO_JSON_AS_NUM(j_mem_info, meminfo_names[i].in_json, memsize);
					break;
				}
			}
		}
		close(fd);
	}

	struct name_name status_names[] = {{"VmPeak:", K_MI_VMPEAK},
	                                   {"VmSize:", K_MI_VMSIZE},
	                                   {"VmLck:",  K_MI_VMLCK},
	                                   {"VmPin:",  K_MI_VMPIN},
	                                   {"VmHWM:",  K_MI_VMHWM},
	                                   {"VmRSS:",  K_MI_VMRSS},
	                                   {"VmData:", K_MI_VMDATA},
	                                   {"VmStk:",  K_MI_VMSTK},
	                                   {"VmExe:",  K_MI_VMEXE},
	                                   {"VmLib:",  K_MI_VMLIB},
	                                   {"VmPTE:",  K_MI_VMPTE},
	                                   {"VmSwap:", K_MI_VMSWAP}};

	snprintf(file_path, PATH_MAX, "/proc/%d/status", pid);
	if ((fd = open(file_path, O_RDONLY)) < 0) {
		_E("open(%s): %m", file_path);
	} else {
		while (fgets_fd(linebuf, BUF_SIZE, fd) != NULL) {
			sscanf(linebuf, "%16s %16s %*s", infoname, memsize);
			for (size_t i = 0; i < ARRAY_SIZE(status_names); i++) {
				if (strcmp(status_names[i].in_file, infoname) == 0) {
					ADD_TO_JSON_AS_NUM(j_mem_info, status_names[i].in_json, memsize);
					break;
				}
			}
		}
		close(fd);
	}

	json_object_object_add(jobj, K_MEM_INFO, j_mem_info);
#undef ADD_TO_JSON_AS_NUM
}

/**
 * @brief Main function.
 *
 * Main module accepts should be launched with:
 *
 *     crash-stack --pid pid [--tid tid] [--sig sig]
 *
 * It allows connecting to a live process and displaying its call stack.
 * It might be also used for connecting to a process from system's core dump handler.
 */
int main(int argc, char **argv)
{
	FILE *outputfile = NULL;
	FILE *jsonoutputfile = NULL;
	int prstatus_fd = -1;
	int c;
	int signo = 0;
	pid_t pid = 0;
	pid_t tid = 0;
	char *p = NULL;
	json_object *j_root = json_object_new_object();
	json_object_object_add(j_root, K_REP_VER, json_object_new_int(INFO_JSON_VERSION));

	while ((c = getopt_long_only(argc, argv, "", opts, NULL)) != -1) {
		switch (c) {
		case OPT_PID:
			pid = atoi(optarg);
			break;
		case OPT_TID:
			tid = atoi(optarg);
			break;
		case OPT_SIGNUM:
			signo = atoi(optarg);
			break;
		case OPT_OUTPUTFILE:
			outputfile = fopen(optarg, "w");
			break;
		case OPT_JSONOUTPUTFILE:
			jsonoutputfile = fopen(optarg, "w");
			break;
		case OPT_PRSTATUS_FD:
			prstatus_fd = strtol(optarg, &p, 10);
			if (*p != 0)
				prstatus_fd = -1;
			break;
		}
	}

	if (NULL == outputfile) outputfile = stdout;

	if (tid == 0) {
		_E("TID not provided");
		return errno;
	}

	argc -= optind;

	/* Executable File Path */
	__crash_stack_print_exe(j_root, pid);

	/* Signal information */
	__crash_stack_print_signal(j_root, signo);

	/* Memory information */
	__crash_stack_print_meminfo(j_root, pid);

	/* Threads */
	__crash_stack_print_threads(j_root, pid, tid);

	/* Maps information */
	__crash_stack_print_maps(j_root, pid);

	if (pid <= 1) {
		_E( "Usage: %s [--output file] [--erroutput file] [--output-json file] [--pid <pid> [--tid <tid>]]",
				argv[0]);

		return 1;
	}

	if (-1 == __get_registers_fd(tid, prstatus_fd))
		return 3333;

	/* Unwind call stack */
	Callstack callstack;
	callstack_constructor(&callstack);

	_create_crash_stack(NULL, tid, &callstack,
			    _crash_stack_get_memory_for_registers(NULL));

	/* Print registers */
	_crash_stack_print_regs(j_root);

	/* Print the results */
	__print_callstack(&callstack, j_root, tid);

	if (jsonoutputfile != NULL) {
		fprintf(jsonoutputfile, "%s", json_object_to_json_string_ext(j_root, JSON_C_TO_STRING_PRETTY));
		fclose(jsonoutputfile);
	}

	if (outputfile != NULL) {
		print_report_info(j_root, outputfile);
		fclose(outputfile);
	}

	/* Clean up */
	callstack_destructor(&callstack);
	json_object_put(j_root);

	return 0;
}
